package java.lang.ref;

import java.util.function.Consumer;
import sun.misc.VM;

public class ReferenceQueue<T>
{
    static ReferenceQueue<Object> NULL;
    static ReferenceQueue<Object> ENQUEUED;
    private ReferenceQueue.Lock lock;
    private volatile Reference<? extends T> head;
    private long queueLength;
    
    public ReferenceQueue() {
        this.lock = new ReferenceQueue.Lock();
        this.head = null;
        this.queueLength = 0L;
    }
    
    boolean enqueue(Reference<? extends T> head) {
        synchronized (this.lock) {
            ReferenceQueue<Object> queue = head.queue;
            if (queue == ReferenceQueue.NULL || queue == ReferenceQueue.ENQUEUED) {
                return false;
            }
            assert queue == this;
            head.queue = ReferenceQueue.ENQUEUED;
            head.next = ((this.head == null) ? head : this.head);
            this.head = head;
            ++this.queueLength;
            if ((head instanceof FinalReference)) {
                VM.addFinalRefCount(1);
            }
            this.lock.notifyAll();
            return true;
        }
    }
    
    private Reference<? extends T> reallyPoll() {
        Reference<Object> head = this.head;
        if (head != null) {
            Reference<Object> next = head.next;
            this.head = ((next == head) ? null : next);
            head.queue = ReferenceQueue.NULL;
            head.next = head;
            --this.queueLength;
            if ((head instanceof FinalReference)) {
                VM.addFinalRefCount(-1);
                return head;
            }
            return head;
        }
        else {
            return null;
        }
    }
    
    public Reference<? extends T> poll() {
        if (this.head == null) {
            return null;
        }
        synchronized (this.lock) {
            return this.reallyPoll();
        }
    }
    
    public Reference<? extends T> remove(long timeoutMillis) throws IllegalArgumentException, InterruptedException {
        if (timeoutMillis < 0L) {
            throw new IllegalArgumentException("Negative timeout value");
        }
        synchronized (this.lock) {
            Reference<Object> reallyPoll = this.reallyPoll();
            if (reallyPoll != null) {
                return reallyPoll;
            }
            long n = (timeoutMillis == 0L) ? 0L : System.nanoTime();
            while (true) {
                this.lock.wait(timeoutMillis);
                Reference<Object> reallyPoll2 = this.reallyPoll();
                if (reallyPoll2 != null) {
                    return reallyPoll2;
                }
                if (timeoutMillis == 0L) {
                    continue;
                }
                long nanoTime = System.nanoTime();
                timeoutMillis -= (nanoTime - n) / 1000000L;
                if (timeoutMillis <= 0L) {
                    return null;
                }
                n = nanoTime;
            }
        }
    }
    
    public Reference<? extends T> remove() throws InterruptedException {
        return this.remove(0L);
    }
    
    void forEach(Consumer<? super Reference<? extends T>> consumer) {
        Reference<Object> reference = this.head;
        while (reference != null) {
            consumer.accept(reference);
            Reference<Object> next = reference.next;
            if (next == reference) {
                if (reference.queue == ReferenceQueue.ENQUEUED) {
                    reference = null;
                }
                else {
                    reference = this.head;
                }
            }
            else {
                reference = next;
            }
        }
    }
    
    static {
        ReferenceQueue.NULL = new ReferenceQueue.Null();
        ReferenceQueue.ENQUEUED = new ReferenceQueue.Null();
    }
    
    private static class Null<S> extends ReferenceQueue<S>
    {
        @Override
        boolean enqueue(Reference<? extends S> reference) {
            return false;
        }
    }
    
    private static class Lock
    {
    }
}
